@@ -35,6 +35,8 @@ import           Data.Maybe                    (catMaybes)
 import           Data.Semigroup                ((<>))
 #endif
 import qualified Data.Vector                   as V
+import qualified Data.Text.Encoding            as Text
+import qualified Data.Text                     as Text
 import           GHC.IO.IOMode                 (IOMode (..))
 import           Numeric                       (showHex)
 import           RIO.ByteString                as B (ByteString, hGetSome, null)
@@ -63,6 +65,8 @@ data IfaceVersion
   | V8101
   | V9001
   | V9041
+  | V9045
+  | V9081
   deriving (Show,Eq,Ord,Enum)
   -- careful, the Ord matters!
 
@@ -205,6 +209,10 @@ getDictionary ptr = do
     traceGet ("Dictionary: " ++ show dict)
     return dict
 
+-- | Get a FastString
+--
+-- FastStrings are stored in a global FastString table and only the index (a
+-- Word32be) is stored at the expected position.
 getCachedBS :: Dictionary -> Get ByteString
 getCachedBS d = go =<< traceShow "Dict index:" getWord32be
   where
@@ -515,15 +523,19 @@ getInterfaceRecent version d = do
     dusage <- traceShow "Usage:"        getUsage
     pure (Interface ddeps dusage)
   where
+    since v = when (version >= v)
+
+    getFastString = getCachedBS d
+
     getModule = do
         idType <- traceShow "Unit type:" getWord8
         case idType of
-            0 -> void $ getCachedBS d
+            0 -> void getFastString
             1 ->
                 void $
-                getCachedBS d *> getList (getTuple (getCachedBS d) getModule)
+                getFastString *> getList (getTuple getFastString getModule)
             _ -> fail $ "Invalid unit type: " <> show idType
-        Module <$> getCachedBS d
+        Module <$> getFastString
     getDependencies =
         withBlockPrefix $ do
           if version >= V9041
@@ -532,22 +544,22 @@ getInterfaceRecent version d = do
               -- only direct imports!
               -- Modules are now prefixed with their UnitId (should have been
               -- ModuleWithIsBoot...)
-              direct_mods <- traceShow "direct_mods:" $ getList (getCachedBS d *> getTuple (getCachedBS d) getBool)
-              direct_pkgs <- getList (getCachedBS d)
+              direct_mods <- traceShow "direct_mods:" $ getList (getFastString *> getTuple getFastString getBool)
+              direct_pkgs <- getList getFastString
 
               -- plugin packages are now stored separately
-              plugin_pkgs <- getList (getCachedBS d)
+              plugin_pkgs <- getList getFastString
               let all_pkgs = unList plugin_pkgs ++ unList direct_pkgs
 
               -- instead of a trust bool for each unit, we have an additional
               -- list of trusted units (transitive)
-              trusted_pkgs <- getList (getCachedBS d)
+              trusted_pkgs <- getList getFastString
               let trusted u = u `elem` unList trusted_pkgs
               let all_pkgs_trust = List (zip all_pkgs (map trusted all_pkgs))
 
               -- these are new
               _sig_mods  <- getList getModule
-              _boot_mods <- getList (getCachedBS d *> getTuple (getCachedBS d) getBool)
+              _boot_mods <- getList (getFastString *> getTuple getFastString getBool)
 
               dep_orphs  <- getList getModule
               dep_finsts <- getList getModule
@@ -557,49 +569,54 @@ getInterfaceRecent version d = do
 
               pure (Dependencies direct_mods all_pkgs_trust dep_orphs dep_finsts dep_plgins)
             else do
-              dep_mods   <- getList (getTuple (getCachedBS d) getBool)
-              dep_pkgs   <- getList (getTuple (getCachedBS d) getBool)
+              dep_mods   <- getList (getTuple getFastString getBool)
+              dep_pkgs   <- getList (getTuple getFastString getBool)
               dep_orphs  <- getList getModule
               dep_finsts <- getList getModule
-              dep_plgins <- getList (getCachedBS d)
+              dep_plgins <- getList getFastString
               pure (Dependencies dep_mods dep_pkgs dep_orphs dep_finsts dep_plgins)
 
     getUsage = withBlockPrefix $ List . catMaybes . unList <$> getList go
       where
+        -- this must follow the `Binary Usage` instance in GHC
+        -- (in GHC.Unit.Module.Deps, at least in GHC 9.4.5)
         go :: Get (Maybe Usage)
         go = do
             usageType <- traceShow "Usage type:" getWord8
             case usageType of
                 0 -> do
-                  void (traceShow "Module:" getModule)
-                  void getFP
-                  void getBool
+                  void (traceShow "Module:" getModule) -- usg_mod
+                  void getFP                           -- usg_mod_hash
+                  void getBool                         -- usg_safe
                   pure Nothing
 
                 1 -> do
-                    void (traceShow "Home module:" (getCachedBS d))
-                    void getFP
-                    void (getMaybe getFP)
-                    void (getList (getTuple (getWord8 *> getCachedBS d) getFP))
-                    void getBool
+                    void (traceShow "Home module:" getFastString)   -- usg_mod_name
+                    since V9045 $ void getFastString                -- usg_unit_id
+                    void getFP                                      -- usg_mod_hash
+                    void (getMaybe getFP)                           -- usg_exports
+                    void (getList (getTuple (getWord8 *> getFastString) getFP)) -- usg_entities
+                    void getBool                                    -- usg_safe
                     pure Nothing
 
                 2 -> do
-                  file_path  <- traceShow "File:" getString
-                  _file_hash <- traceShow "FP:" getFP'
-                  when (version >= V9041) $ do
-                    _file_label <- traceShow "File label:" (getMaybe getString)
-                    pure ()
+                  -- usg_file_path
+                  file_path  <- traceShow "File:" $ if version >= V9081
+                    then Text.unpack . Text.decodeUtf8 <$> getFastString
+                    else getString
+                  void $ traceShow "FP:" getFP'                     -- usg_file_hash
+                  since V9041 $ void $ traceShow "File label:" (getMaybe getString)-- usg_file_label
                   pure (Just (Usage file_path))
 
                 3 -> do
-                  void getModule
-                  void getFP
+                  void getModule -- usg_mod
+                  void getFP     -- usg_mod_hash
                   pure Nothing
 
                 4 | version >= V9041 -> do -- UsageHomeModuleInterface
-                  _mod_name   <- void (getCachedBS d)
-                  _iface_hash <- void getFP
+                  void getFastString                  -- usg_mod_name
+                  since V9045 $ void getFastString    -- usg_unit_id
+                  void getFP                          -- usg_iface_hash
                   pure Nothing
 
                 _ -> fail $ "Invalid usageType: " <> show usageType
@@ -641,6 +658,8 @@ getInterface = do
     traceGet ("Version: " ++ version)
 
     let !ifaceVersion
+          | version >= "9081" = V9081
+          | version >= "9045" = V9045
           | version >= "9041" = V9041
           | version >= "9001" = V9001
           | version >= "8101" = V8101
@@ -675,6 +694,8 @@ getInterface = do
     void getPtr
 
     case ifaceVersion of
+      V9081 -> getInterfaceRecent ifaceVersion dict
+      V9045 -> getInterfaceRecent ifaceVersion dict
       V9041 -> getInterfaceRecent ifaceVersion dict
       V9001 -> getInterfaceRecent ifaceVersion dict
       V8101 -> getInterfaceRecent ifaceVersion dict
